import pendulum

from swimlane.core.resources.base import APIResource
from swimlane.core.resources.usergroup import UserGroup


class RevisionBase(APIResource):
    """
    The base class representing a single revision returned from a History lookup.

    Attributes:
        Attributes:
        modified_date: The date this revision was created.
        revision_number: The revision number of this revision.
        status: Indicates whether this revision is the current revision or a historical revision.
        user: The user that saved this revision.
    """

    def __init__(self, swimlane, raw):
        super(RevisionBase, self).__init__(swimlane, raw)

        self.modified_date = pendulum.parse(self._raw['modifiedDate'])
        self.revision_number = self._raw['revisionNumber']
        self.status = self._raw['status']

        # UserGroupSelection, can't set as User without additional lookup
        self.user = UserGroup(self._swimlane, self._raw['userId'])

        self._raw_version = self._raw['version']
        self._version = None

    def __str__(self):
        return '{} ({})'.format(self.version, self.revision_number)

    @property
    def version(self):
        raise NotImplementedError

    def for_json(self):
        """Return revision metadata"""
        return {
            'modifiedDate': self._raw['modifiedDate'],
            'revisionNumber': self.revision_number,
            'user': self.user.for_json()
        }



from swimlane.core.resources.app import App
from swimlane.core.resources.revision_base import RevisionBase


class AppRevision(RevisionBase):
    """
    Encapsulates a single revision returned from a History lookup.

    Attributes:
        Attributes:
        modified_date: The date this app revision was created.
        revision_number: The revision number of this app revision.
        status: Indicates whether this revision is the current revision or a historical revision.
        user: The user that saved this revision of the record.
        version: The App corresponding to the data contained in this app revision.
    """

    # Separator for unique ids. Unlikely to be found in application ids. Although technically we do not currently
    # validate app ids in the backend for specific characters so this sequence could be found.
    SEPARATOR = ' --- '

    @staticmethod
    def get_unique_id(app_id, revision_number):
        """Returns the unique identifier for the given AppRevision."""
        return '{0}{1}{2}'.format(app_id, AppRevision.SEPARATOR, revision_number)

    @staticmethod
    def parse_unique_id(unique_id):
        """Returns an array containing two items: the app_id and revision number parsed from the given unique_id."""
        return unique_id.split(AppRevision.SEPARATOR)

    @property
    def version(self):
        """Returns an App from the _raw_version info in this app revision. Lazy loaded. Overridden from base class."""
        if not self._version:
            self._version = App(self._swimlane, self._raw_version)
        return self._version

    def get_cache_index_keys(self):
        """Returns cache index keys for this AppRevision."""
        return {
            'app_id_revision': self.get_unique_id(self.version.id, self.revision_number)
        }



from swimlane.core.resolver import SwimlaneResolver
from swimlane.core.cache import check_cache
from swimlane.core.resources.task import Task
from swimlane.utils import one_of_keyword_only


class TaskAdapter(SwimlaneResolver):
    """Handles retreival of Swimlane Task Resources and execution of tasks against records."""
    
    @check_cache(Task)
    @one_of_keyword_only('id', 'name')
    def get(self, key, value):
        """Get a single task by id or name"""
        if key == 'id':
            response = self._swimlane.request('get', 'task/{id}'.format(id=value))
            return Task(self._swimlane, response.json())

        if key == 'name':
            response = self._swimlane.request('get', 'task/light')
            for task in response.json():
                if value == task.get('name'):
                    return self._get_full(task.get('id'))
            raise ValueError('No task with name "{value}"'.format(value=value))

    def list(self):
        """Retrieve list of all tasks.

        Returns:
            :class:`list` of :class:`~swimlane.core.resources.task.Task`: List of all tasks.
        """
        response = self._swimlane.request('get', 'task/light')
        return [self._get_full(item.get('id')) for item in response.json()]
    
    def execute(self, task_name, raw_record):
        """Execute job by name.
        
        Returns:
            Response of request from the API endpoint.
        """
        task_id = self.get(name=task_name).id
        data = {
            'taskId': task_id,
            'record': raw_record,
        }
        return self._swimlane.request('post', 'task/execute/record', json=data)
    
    def _get_full(self, task_id):
        """Retreived complete task raw json of task from API.

        Returns:
            :class:`~swimlane.core.resources.task.Task`: Single Task object.
        """
        response = self._swimlane.request('get', 'task/{task_id}'.format(task_id=task_id))
        return Task(self._swimlane, response.json())



"""Custom exceptions and errors"""

from difflib import get_close_matches

from requests import HTTPError


class SwimlaneException(Exception):
    """Base exception for Swimlane errors"""


class UnknownField(SwimlaneException, KeyError):
    """Raised anytime access is attempted to a field that does not exist on an App or Record

    Attributes:
        app (App): App with the unknown field requested
        field_name (str): Name of the field that was requested
        similar_field_names (list(str)): List of strings of fields on app that are potentially similar to field_name
    """

    def __init__(self, app, field_name, field_pool):
        self.app = app
        self.field_name = field_name
        self.similar_field_names = get_close_matches(self.field_name, field_pool, 3)

        message = "{!r} has no field '{}'".format(self.app, self.field_name)

        if self.similar_field_names:
            message += '. Similar fields: ' + ', '.join([repr(f) for f in self.similar_field_names])

        super(UnknownField, self).__init__(message)


class ValidationError(SwimlaneException, ValueError):
    """Raised when record's field data is invalid

    Attributes:
        record (Record): Record in context of validation failure
        failure (str): Reason for record failure
    """

    def __init__(self, record, failure):
        self.record = record
        self.failure = failure

        super(ValidationError, self).__init__(
            'Validation failed for {!r}. Reason: {}'.format(self.record, self.failure)
        )


class _InvalidSwimlaneVersion(SwimlaneException, NotImplementedError):
    """Base class raised when connecting to unsupported versions of Swimlane

    Attributes:
        swimlane (Swimlane): Swimlane client failing the version check
        min_version (str): Minimum version specified on version range
        max_version (str): Maximum version specified on version range
    """

    def __init__(self, swimlane, min_version, max_version):
        self.swimlane = swimlane
        self.min_version = min_version
        self.max_version = max_version

        super(_InvalidSwimlaneVersion, self).__init__(self._get_message())

    def _get_range_string(self):
        if self.min_version and self.max_version:
            range_string = '>= {}, < {}'.format(self.min_version, self.max_version)
        elif self.min_version:
            range_string = '>= {}'.format(self.min_version)
        else:
            range_string = '<= {}'.format(self.max_version)

        return range_string

    def _get_message(self):
        return 'Swimlane version {}, must be {}'.format(
            self.swimlane.version,
            self._get_range_string()
        )


class InvalidSwimlaneBuildVersion(_InvalidSwimlaneVersion):
    """Raised when method connected to Swimlane with build version outside a required range"""

    def _get_message(self):
        return 'Swimlane build version {}, must be {}'.format(
            self.swimlane.build_version,
            self._get_range_string()
        )


class InvalidSwimlaneProductVersion(_InvalidSwimlaneVersion):
    """Raised when method connected to Swimlane with product version outside a required range"""

    def _get_message(self):
        return 'Swimlane product version {}, must be {}'.format(
            self.swimlane.product_version,
            self._get_range_string()
        )


class SwimlaneHTTP400Error(SwimlaneException, HTTPError):
    """Exception raised when receiving a 400 response with additional context

    Attributes:
        code (int): Swimlane error code
        name (str): Human-readable Swimlane error name
        argument (str): Optional argument included with error or None
        http_error (HTTPError): Source requests.HTTPError caught and used to generate this exception
    """

    codes = {
        -1: 'Unknown',
        1000: 'PasswordExpired',
        1001: 'DuplicateUserName',
        1002: 'InvalidUserNameOrPassword',
        1003: 'ConfirmPasswordDoesNotMatch',
        1004: 'PasswordDoesNotMeetComplexityRequirements',
        1005: 'PasswordResetRequired',
        1006: 'NewPasswordCannotMatchCurrent',
        1007: 'InvalidUser',
        1051: 'DuplicateGroupName',
        1061: 'DuplicateRoleName',
        2000: 'DuplicateFieldName',
        2001: 'FieldNameEmpty',
        2002: 'InvalidApplicationExportFile',
        2003: 'ApplicationNotFound',
        2004: 'InvalidCalculation',
        2005: 'DuplicateApplicationName',
        2006: 'DuplicateAppletName',
        2007: 'DuplicateAppletAcronym',
        2008: 'DuplicateApplicationAcronym',
        2011: 'SectionNameTooLong',
        3000: 'DuplicateFieldValue',
        3001: 'InvalidDateField',
        3002: 'RecordNotFound',
        3003: 'FieldNotFound',
        3006: 'MaxAttachmentSize',
        4000: 'BadStatsGroup',
        4001: 'BadFilter',
        5000: 'AppLimitExceeded',
        5001: 'UserLimitExceeded',
        5002: 'NewServerInstall',
        5003: 'UnableToConnectToActiveDirectory',
        5004: 'UnableToRetrieveStoredValue',
        5005: 'UnableToConnectToMongoDb',
        5006: 'UnableToConnectToSmtp',
        5007: 'SwimlaneAlreadyInitialized',
        5008: 'ModelValidationError',
        5009: 'UpgradeInProcess',
        5010: 'RequiredFieldMissing',
        5011: 'UnableToRetrieveEncryptionKey',
        5012: 'PathNotFound',
        5013: 'WrongType',
        5014: 'ModificationError',
        5015: 'DatabaseError',
        5016: 'NetworkError',
        5017: 'InvalidOnThisOS',
        6000: 'ConnectionDataNotProvided',
        7000: 'RegexNotDefined',
        7001: 'AssetNotFound',
        9000: 'BadThreatIntelConnector',
        9001: 'NoThreatIntel',
        9002: 'ThreatIntelTypeNotSupportedByThisProvider',
        10000: 'DuplicateTaskName',
        10001: 'TaskNotFound',
        17001: 'DuplicateAssetName',
        19001: 'HangfireError'
    }

    def __init__(self, http_error):
        self.http_error = http_error

        try:
            error_content = self.http_error.response.json()
        except ValueError:
            error_content = {'Argument': None, 'ErrorCode': '-1'}

        self.code = int(error_content.get('ErrorCode', -1))
        self.argument = error_content.get('Argument')
        self.name = self.codes.get(self.code, self.codes[-1])

        message = '{}:{}'.format(self.name, self.code)
        if self.argument is not None:
            message = '{message} ({argument})'.format(message=message, argument=self.argument)

        super(SwimlaneHTTP400Error, self).__init__(
            '{message}: Bad Request for url: {url}'.format(message=message, url=self.http_error.response.url)
        )



import six

from swimlane.core.resolver import SwimlaneResolver


class APIResourceMetaclass(type):
    """Metaclass for all APIResource classes"""

    def __call__(cls, *args, **kwargs):
        """Hook __init__ call to push resource instance into Swimlane client ResourceCache after instantiation"""
        resource_instance = type.__call__(cls, *args, **kwargs)

        resource_instance._swimlane.resources_cache.cache(resource_instance)

        return resource_instance


class APIResource(six.with_metaclass(APIResourceMetaclass, SwimlaneResolver)):
    """Base class for all API resources with an associated $type and/or raw data"""

    _type = None

    def __init__(self, swimlane, raw):
        super(APIResource, self).__init__(swimlane)
        self._raw = raw

        raw_type = self._raw.get('$type')
        if self._type and raw_type != self._type:
            raise TypeError('Expected $type = "{}", received "{}"'.format(self._type, raw_type))

    def __repr__(self):
        return '<{self.__class__.__name__}: {self!s}>'.format(self=self)

    def __str__(self):
        return ''

    def __hash__(self):
        """Added for py2+3 compat"""
        return int(id(self) / 16)

    def __eq__(self, other):
        """Determine if an APIResource is of the same type and has the same hash value"""
        return isinstance(other, self.__class__) and hash(self) == hash(other)

    def __ne__(self, other):
        # Default __ne__ for python 2 compat
        return not self == other

    def get_cache_internal_key(self):
        """Return real internal cache key for resource instance"""
        return hash(self)

    def get_cache_index_keys(self):
        """Return dict of key/value pairs used by ResourceCache to map resource values to internal cache instance"""
        raise NotImplementedError



from swimlane.core.resolver import AppResolver
from swimlane.core.resources.record_revision import RecordRevision


class RecordRevisionAdapter(AppResolver):
    """Handles retrieval of Swimlane Record Revision resources"""

    def __init__(self, app, record):
        super(RecordRevisionAdapter, self).__init__(app)
        self.record = record

    def get_all(self):
        """Get all revisions for a single record.

        Returns:
            RecordRevision[]: All record revisions for the given record ID.
        """
        raw_revisions = self._swimlane.request('get',
                                               'app/{0}/record/{1}/history'.format(self._app.id, self.record.id)).json()
        return [RecordRevision(self._app, raw) for raw in raw_revisions]

    def get(self, revision_number):
        """Gets a specific record revision.

        Keyword Args:
            revision_number (float): Record revision number

        Returns:
            RecordRevision: The RecordRevision for the given revision number.
        """
        record_revision_raw = self._swimlane.request('get',
                                                     'app/{0}/record/{1}/history/{2}'.format(self._app.id,
                                                                                             self.record.id,
                                                                                             revision_number)).json()
        return RecordRevision(self._app, record_revision_raw)



import pendulum

from swimlane.core.cursor import PaginatedCursor
from swimlane.core.resources.base import APIResource
from swimlane.core.resources.record import Record, record_factory
from swimlane.core.search import CONTAINS, EQ, EXCLUDES, NOT_EQ, LT, GT, LTE, GTE, ASC, DESC


class Report(APIResource, PaginatedCursor):
    """A report class used for searching

    Can be iterated over to retrieve results

    Notes:
        Record retrieval is lazily evaluated and cached internally, adding a filter and attempting to iterate again will
        not respect the additional filter and will return the same set of records each time

    Examples:

        Lazy retrieval of records with direct iteration over report

        ::

            report = app.reports.build('new-report')
            report.filter('field_1', 'equals', 'value')

            for record in report:
                do_thing(record)

        Full immediate retrieval of all records

        ::

            report = app.reports.build('new-report')
            report.filter('field_1', 'doesNotEqual', 'value')

            records = list(report)


    Attributes:
        name (str): Report name

    Keyword Args:
        limit (int): Max number of records to return from report/search
        page_size (int): Max number of records per page
        keywords (list(str)): List of keywords to use in report/search
    """

    _type = "Core.Models.Search.Report, Core"

    _FILTER_OPERANDS = (
        EQ,
        NOT_EQ,
        CONTAINS,
        EXCLUDES,
        LT,
        GT,
        LTE,
        GTE
    )

    _SORT_ORDERS = (
        ASC,
        DESC
    )

    default_limit = 50

    def __init__(self, app, raw, **kwargs):
        APIResource.__init__(self, app._swimlane, raw)
        PaginatedCursor.__init__(self,
                                 limit=kwargs.pop('limit', self.default_limit),
                                 page_size=kwargs.pop('page_size', self.default_page_size))

        self.name = self._raw['name']
        self.keywords = kwargs.pop('keywords', [])

        self._app = app

        for field_id in self._app._fields_by_id.keys():
            self._raw['columns'].append(field_id)

    def __str__(self):
        return self.name

    def _retrieve_raw_elements(self, page):
        body = self._raw.copy()

        body['pageSize'] = self.page_size
        body['offset'] = page
        body['keywords'] = ', '.join(self.keywords)

        response = self._swimlane.request('post', 'search', json=body)
        return response.json()['results'].get(self._app.id, [])

    def _parse_raw_element(self, raw_element):
        return Record(self._app, raw_element)

    def filter(self, field_name, operand, value):
        """Adds a filter to report

        Notes:
            All filters are currently AND'ed together

        Args:
            field_name (str): Target field name to filter on
            operand (str): Operand used in comparison. See `swimlane.core.search` for options
            value: Target value used in comparison
        """
        if operand not in self._FILTER_OPERANDS:
            raise ValueError('Operand must be one of {}'.format(', '.join(self._FILTER_OPERANDS)))

        field = self._get_stub_field(field_name)

        self._raw['filters'].append({
            "fieldId": field.id,
            "filterType": operand,
            "value": field.get_report(value)
        })

    def sort(self, field_name, order):
        """Adds a sort to report

        Args:
            field_name (str): Target field name to sort by
            order (str): Sort order
        """
        if (order not in self._SORT_ORDERS):
            raise ValueError('Order must be one of {}'.format(', '.join(self._SORT_ORDERS)))

        field = self._get_stub_field(field_name)

        self._raw['sorts'][field.id] = order

    def set_columns(self, *field_names):
        """Set specified columns for report

        Notes:
            The Tracking Id column is always included

        Args:
            *field_names (str): Zero or more column names
        """
        self._raw['columns'] = []
        for field_name in field_names:
            field = self._get_stub_field(field_name)

            self._raw['columns'].append(field.id)

        if self._app.tracking_id not in self._raw['columns']:
            self._raw['columns'].append(self._app.tracking_id)

    def _get_stub_field(self, field_name):
        # Use temp Record instance for target app to translate values into expected API format
        record_stub = record_factory(self._app)
        return record_stub.get_field(field_name)

def report_factory(app, report_name, **kwargs):
    """Report instance factory populating boilerplate raw data

    Args:
        app (App): Swimlane App instance
        report_name (str): Generated Report name

    Keyword Args
        **kwargs: Kwargs to pass to the Report class
    """
    # pylint: disable=protected-access
    created = pendulum.now().to_rfc3339_string()
    user_model = app._swimlane.user.as_usergroup_selection()

    return Report(
        app,
        {
            "$type": Report._type,
            "groupBys": [],
            "aggregates": [],
            "applicationIds": [app.id],
            "columns": [],
            "sorts": {
                "$type": "System.Collections.Generic.Dictionary`2"
                         "[[System.String, mscorlib],"
                         "[Core.Models.Search.SortTypes, Core]], mscorlib",
            },
            "filters": [],
            "defaultSearchReport": False,
            "allowed": [],
            "permissions": {
                "$type": "Core.Models.Security.PermissionMatrix, Core"
            },
            "createdDate": created,
            "modifiedDate": created,
            "createdByUser": user_model,
            "modifiedByUser": user_model,
            "id": None,
            "name": report_name,
            "disabled": False,
            "keywords": ""
        },
        **kwargs
    )



import requests

class WrappedSession(requests.Session):
    """A wrapper for requests.Session to override 'verify' property, ignoring REQUESTS_CA_BUNDLE environment variable.

    This is a workaround for https://github.com/kennethreitz/requests/issues/3829 (will be fixed in requests 3.0.0)
    """
    def merge_environment_settings(self, url, proxies, stream, verify, *args, **kwargs):
        if self.verify is False:
            verify = False

        return super(WrappedSession, self).merge_environment_settings(url, proxies, stream, verify, *args, **kwargs)


import functools
import re

from pkg_resources import get_distribution, DistributionNotFound

from swimlane.exceptions import InvalidSwimlaneBuildVersion


def compare_versions(version_a, version_b, zerofill=False):
    """Return direction of version relative to provided version sections

    Args:
        version_a (str): First version to compare
        version_b (str): Second version to compare
        zerofill (bool): If True, treat any missing version sections as 0, otherwise ignore section. Defaults to False

    Returns:
        int: 0 if equal, -1 if a > b, 1 if a < b

    Examples:

        If a is equal to b, return 0
        If a is greater than b, return -1
        If a is less than b, return 1

        >>> compare_versions('2', '2') == 0
        >>> compare_versions('2', '1') == -1
        >>> compare_versions('2', '3') == 1

        If zerofill is False (default), sections not included in both versions are ignored during comparison

        >>> compare_versions('2.13.2', '2.13') == 0
        >>> compare_versions('2.13.2-1234', '3') == 1

        If zerofill is True, any sections in one version not included in other version are set to 0

        >>> compare_versions('2.13.2', '2.13', True) == -1
        >>> compare_versions('2.13.2-1234', '2.13.2', True) == -1
        >>> compare_versions('2.13.2', '2.13.2', True) == 0
    """
    a_sections = list((int(match) for match in re.findall(r'\d+', version_a)))
    b_sections = list((int(match) for match in re.findall(r'\d+', version_b)))

    if zerofill:
        max_sections = max([len(a_sections), len(b_sections)])

        a_sections += [0 for _ in range(max(max_sections - len(a_sections), 0))]
        b_sections += [0 for _ in range(max(max_sections - len(b_sections), 0))]

    else:
        min_sections = min([len(a_sections), len(b_sections)])

        a_sections = a_sections[:min_sections]
        b_sections = b_sections[:min_sections]

    return (b_sections > a_sections) - (b_sections < a_sections)


def requires_swimlane_version(min_version=None, max_version=None):
    """Decorator for SwimlaneResolver methods verifying Swimlane server build version is within a given inclusive range

    Raises:
        InvalidVersion: Raised before decorated method call if Swimlane server version is out of provided range
        ValueError: If neither min_version or max_version were provided, or if those values conflict (2.15 < 2.14)
    """

    if min_version is None and max_version is None:
        raise ValueError('Must provide either min_version, max_version, or both')

    if min_version and max_version and compare_versions(min_version, max_version) < 0:
        raise ValueError('min_version must be <= max_version ({}, {})'.format(min_version, max_version))

    def decorator(func):

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            swimlane = self._swimlane

            if min_version and compare_versions(min_version, swimlane.build_version, True) < 0:
                raise InvalidSwimlaneBuildVersion(swimlane, min_version, max_version)

            if max_version and compare_versions(swimlane.build_version, max_version, True) < 0:
                raise InvalidSwimlaneBuildVersion(swimlane, min_version, max_version)

            return func(self, *args, **kwargs)

        return wrapper

    return decorator


def get_package_version():
    """Get swimlane package version

    Returns:
         str: Installed swimlane lib package version, or 0.0.0.dev if not fully installed
    """
    try:
        return get_distribution(__name__.split('.')[0]).version
    except DistributionNotFound:
        return '0.0.0.dev'



from .base import CursorField, FieldCursor, ReadOnly


class RevisionCursor(FieldCursor):
    """An iterable object that automatically lazy retrieves and caches history data for a record from API"""

    def __init__(self, *args, **kwargs):
        super(RevisionCursor, self).__init__(*args, **kwargs)
        self.__retrieved = False

    def _evaluate(self):
        """Lazily retrieves, caches, and returns the list of record _revisions"""
        if not self.__retrieved:
            self._elements = self._retrieve_revisions()
            self.__retrieved = True

        return super(RevisionCursor, self)._evaluate()

    def _retrieve_revisions(self):
        """Populate RecordRevision instances."""
        return self._record.revisions.get_all()


class HistoryField(ReadOnly, CursorField):

    field_type = 'Core.Models.Fields.History.HistoryField, Core'
    cursor_class = RevisionCursor
    bulk_modify_support = False



"""Base classes used to build field abstractions"""
from .cursor import CursorField, FieldCursor
from .field import Field
from .multiselect import MultiSelectField, MultiSelectCursor


class ReadOnly(Field):
    """Mixin explicitly disabling setting value via python"""

    def __init__(self, *args, **kwargs):
        super(ReadOnly, self).__init__(*args, **kwargs)

        self.readonly = True





"""Abstractions for Swimlane app field types to simplify getting/setting values on records"""

from six import string_types as _string_types

from swimlane.core.fields.base import Field
from swimlane.utils import (
    get_recursive_subclasses as _get_recursive_subclasses,
    import_submodules as _import_submodules
)

_import_submodules(__name__)


def _build_field_type_map(base_class):
    """Create mapping from all $type values to their respective Field classes"""
    mapping = {}

    for cls in _get_recursive_subclasses(base_class):
        if cls.field_type:
            if isinstance(cls.field_type, tuple):
                for field_type in cls.field_type:
                    mapping[field_type] = cls
            elif isinstance(cls.field_type, _string_types):
                mapping[cls.field_type] = cls
            else:
                raise ValueError('Field type must be str or tuple, cannot understand type "{}" on class "{}"'.format(
                    type(cls.field_type),
                    cls
                ))

    return mapping


_FIELD_TYPE_MAP = _build_field_type_map(Field)


def resolve_field_class(field_definition):
    """Return field class most fitting of provided Swimlane field definition"""
    try:
        return _FIELD_TYPE_MAP[field_definition['$type']]
    except KeyError as error:
        error.message = 'No field available to handle Swimlane $type "{}"'.format(field_definition)
        raise


__all__ = ['resolve_field_class'] + [f.__class__.__name__ for f in _FIELD_TYPE_MAP.values()]



from functools import total_ordering

import six

from swimlane.exceptions import UnknownField
from .base import APIResource


@total_ordering
class App(APIResource):
    """A single App record instance

    Used lookup field definitions and retrieve/create child Record instances

    Attributes:
        name (str): App name
        acronym (str): App acronym
        description (str): App description
        id (str): Full App ID
        tracking_id (str): App tracking ID
        records (RecordAdapter): :class:`~swimlane.core.adapters.record.RecordAdapter` configured for current App
        reports (ReportAdapter): :class:`~swimlane.core.adapters.report.ReportAdapter` configured for current App
    """

    _type = 'Core.Models.Application.Application, Core'

    def __init__(self, swimlane, raw):
        super(App, self).__init__(swimlane, raw)

        self.acronym = self._raw['acronym']
        self.name = self._raw['name']
        self.description = self._raw.get('description', '')
        self.id = self._raw['id']
        self.tracking_id = self._raw.get('trackingFieldId')

        self._fields_by_id = dict()
        self._fields_by_name = dict()
        self._defaults = dict()

        for field in self._raw['fields']:
            self._fields_by_id[field['id']] = field
            self._fields_by_name[field['name']] = field
            if 'fieldType' in field and field['fieldType'] == "valuesList":
                selection_type = field['selectionType']
                for value in field['values']:
                    if 'selected' in value and value['selected']:
                        if selection_type == 'single':
                            self._defaults[field['name']] = value['name']
                            break
                        else:
                            default = self._defaults.get(field['name'], list())
                            default.extend([value['name']])
                            self._defaults[field['name']] = default

        self._keys_to_field_names = {}
        for name, field_def in six.iteritems(self._fields_by_name):
            # Include original name to simplify name resolution
            self._keys_to_field_names[name] = name
            key = field_def.get('key')
            if key:
                self._keys_to_field_names[key] = name

        # Avoid circular import
        from swimlane.core.adapters import RecordAdapter, ReportAdapter, AppRevisionAdapter
        self.records = RecordAdapter(self)
        self.reports = ReportAdapter(self)
        self.revisions = AppRevisionAdapter(self)

    def __str__(self):
        return '{self.name} ({self.acronym})'.format(self=self)

    def __hash__(self):
        return hash((self.id, self.name))

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("Comparisons not supported between instances of '{}' and '{}'".format(
                other.__class__.__name__,
                self.__class__.__name__
            ))

        return self.name < other.name

    def get_cache_index_keys(self):
        """Return all fields available when retrieving apps"""
        return {
            'id': self.id,
            'name': self.name,
            'acroynm': self.acronym
        }

    def resolve_field_name(self, field_key):
        """Return the field name matching the given key or None. Searches field keys first, falls back to field names"""
        return self._keys_to_field_names.get(field_key)

    def get_field_definition_by_name(self, field_name):
        """Get JSON field definition for field matching provided name or key

        .. versionchanged:: 4.1.0
            Added support for field keys

        Args:
            field_name (str): Target field name or key to get definition for

        Raises:
            swimlane.exceptions.UnknownField: Raised when given a field name not found in App

        Returns:
            dict: Field metadata definition
        """
        try:
            return self._fields_by_name[self.resolve_field_name(field_name)]
        except KeyError:
            raise UnknownField(self, field_name, self._fields_by_name.keys())

    def get_field_definition_by_id(self, field_id):
        """Get JSON field definition for field matching provided id

        Args:
            field_id (str): Target field ID to get definition for

        Raises:
            swimlane.exceptions.UnknownField: Raised when given a field ID not found in App

        Returns:
            dict: Field metadata definition
        """
        try:
            return self._fields_by_id[field_id]
        except KeyError:
            raise UnknownField(self, field_id, self._fields_by_id.keys())



import pendulum

from swimlane.core.resources.base import APIResource
from swimlane.core.resources.usergroup import UserGroup


class Comment(APIResource):
    """Abstraction of a single comment from a comment field

    Attributes:
        user (UserGroup): UserGroup instance of user who created the comment
        created_date (pendulum.DateTime): Pendulum datetime of when comment was created
        message (str): Comment message body
    """

    def __init__(self, swimlane, raw):
        super(Comment, self).__init__(swimlane, raw)

        self.user = UserGroup(swimlane, self._raw['createdByUser'])
        self.created_date = pendulum.parse(self._raw['createdDate'])
        self.message = self._raw['message']
        self.is_rich_text = self._raw.get('isRichText', False)

    def __str__(self):
        return self.message

    def for_json(self):
        """Called by CommentField.for_json(), returns relevant Comment attributes in JSON-compatible format"""
        return {
            'message': self.message,
            'createdDate': self._raw['createdDate'],
            'createdByUser': self.user.for_json()
        }



import mimetypes

from swimlane.core.fields.base import MultiSelectField, FieldCursor
from swimlane.core.resources.attachment import Attachment


class AttachmentCursor(FieldCursor):
    """Allows creation and iteration of attachments"""

    def add(self, filename, stream, content_type=None):
        """Upload a new attachment, and add it to current fields raw data to be persisted on save

        Can optionally manually set the content_type, will be guessed by provided filename extension and default to 
        application/octet-stream if it cannot be guessed
        """
        # Guess file Content-Type or default
        content_type = content_type or mimetypes.guess_type(
            filename)[0] or 'application/octet-stream'
        response = self._record._swimlane.request(
            'post',
            'attachment/{appId}/{fieldId}'.format(
                appId=self._record.app.id, fieldId=self._field.id),
            files={
                'file': (filename, stream, content_type)
            },
        )

        # Returns raw attachment data as list with single element
        raw_attachment_data = response.json()[0]

        attachment = Attachment(self._record._swimlane, raw_attachment_data)
        self._elements.append(attachment)

        self._sync_field()

        return attachment


class AttachmentsField(MultiSelectField):

    field_type = (
        'Core.Models.Fields.AttachmentField, Core',
        'Core.Models.Fields.Attachment.AttachmentField, Core'
    )
    cursor_class = AttachmentCursor
    supported_types = [Attachment]
    bulk_modify_support = False

    def __init__(self, *args, **kwargs):
        """Override to force-set multiselect to always True"""
        super(AttachmentsField, self).__init__(*args, **kwargs)
        self.multiselect = True

    def get_initial_elements(self):
        raw_value = self.get_swimlane() or []

        return [self.cast_to_python(raw) for raw in raw_value]

    def get_batch_representation(self):
        """Return best batch process representation of field value"""
        return self.get_swimlane()

    def _set(self, value):
        """Override setter, allow clearing cursor"""
        super(AttachmentsField, self)._set(value)
        self._cursor = None

    def cast_to_python(self, value):
        return Attachment(self._swimlane, value)

    def cast_to_swimlane(self, value):
        return value._raw



from functools import total_ordering

from swimlane.core.cursor import Cursor
from swimlane.core.resolver import SwimlaneResolver
from swimlane.core.resources.base import APIResource


# pylint: disable=abstract-method
@total_ordering
class UserGroup(APIResource):
    """Base class for Users and Groups

    Notes:
        Returned in some places where determining whether object is a User or Group is not possible without additional
        requests. Use appropriate adapter on `swimlane` client to retrieve more specific instance using `id` as needed

        Can be compared to User or Group instances directly without ensuring the classes are the same

    Attributes:
        id (str): Full user/group ID
        name (str): User/group name
    """

    def __init__(self, swimlane, raw):
        super(UserGroup, self).__init__(swimlane, raw)

        self.id = self._raw['id']
        self.name = self._raw['name']

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        """Override to allow equality comparisons across UserGroup, User, and Group instances"""
        return isinstance(other, UserGroup) and hash(self) == hash(other)

    def __lt__(self, other):
        if not isinstance(other, UserGroup):
            raise TypeError("Comparisons not supported between instances of '{}' and '{}'".format(
                other.__class__.__name__,
                self.__class__.__name__
            ))

        return self.name < other.name

    def resolve(self):
        """Retrieve and return correct User or Group instance from UserGroup

        .. versionadded:: 2.16.1

        Returns:
            User | Group: Resolved User or Group instance
        """
        # Skip resolving if not a generic instance
        if self.__class__ is not UserGroup:
            return self

        else:
            try:
                return self._swimlane.users.get(id=self.id)
            except ValueError:
                return self._swimlane.groups.get(id=self.id)

    def as_usergroup_selection(self):
        """Converts UserGroup to raw UserGroupSelection for populating record

        Returns:
            dict: Formatted UserGroup data as used by selection fields
        """
        return {
            '$type': 'Core.Models.Utilities.UserGroupSelection, Core',
            'id': self.id,
            'name': self.name
        }

    def for_json(self):
        """Get JSON-compatible representation"""
        return {
            'id': self.id,
            'name': self.name
        }


class Group(UserGroup):
    """Swimlane group record

    Attributes:
        description (str): Group description
        users (GroupUsersCursor): List of users belonging to group.
    """

    _type = 'Core.Models.Groups.Group, Core'

    def __init__(self, swimlane, raw):
        super(Group, self).__init__(swimlane, raw)
        self.__user_ids = [item['id'] for item in self._raw.get('users')]
        self.description = self._raw.get('description')
        self.__users = None

    @property
    def users(self):
        """Returns a GroupUsersCursor with list of User instances for this Group

        .. versionadded:: 2.16.2
        """
        if self.__users is None:
            self.__users = GroupUsersCursor(swimlane=self._swimlane, user_ids=self.__user_ids)
        return self.__users

    def get_cache_index_keys(self):
        return {
            'id': self.id,
            'name': self.name
        }


class User(UserGroup):
    """Swimlane user record

    Attributes:
        username (str): Unique username
        display_name (str): User display name
        email (str): User email
    """

    _type = 'Core.Models.Identity.ApplicationUser, Core'

    def __init__(self, swimlane, raw):
        super(User, self).__init__(swimlane, raw)

        self.username = self._raw.get('userName')
        self.display_name = self._raw.get('displayName')
        self.email = self._raw.get('email')

    def get_cache_index_keys(self):
        return {
            'id': self.id,
            'username': self.username,
            'display_name': self.display_name
        }


class GroupUsersCursor(SwimlaneResolver, Cursor):
    """Handles retrieval for user endpoint"""

    def __init__(self, swimlane, user_ids):
        SwimlaneResolver.__init__(self, swimlane)
        Cursor.__init__(self)
        self.__user_ids = user_ids

    def _evaluate(self):
        """Lazily retrieve and build User instances from returned data"""
        if self._elements:
            for element in self._elements:
                yield element
        else:
            for user_id in self.__user_ids:
                element = self._swimlane.users.get(id=user_id)
                self._elements.append(element)
                yield element



from swimlane.core.cache import check_cache
from swimlane.core.resolver import AppResolver
from swimlane.core.resources.app_revision import AppRevision
from swimlane.utils import one_of_keyword_only


class AppRevisionAdapter(AppResolver):
    """Handles retrieval of Swimlane App Revision resources"""

    def get_all(self):
        """
        Gets all app revisions.

        Returns:
            AppRevision[]: Returns all AppRevisions for this Adapter's app.
        """
        raw_revisions = self._swimlane.request('get', 'app/{0}/history'.format(self._app.id)).json()
        return [AppRevision(self._swimlane, raw) for raw in raw_revisions]

    def get(self, revision_number):
        """
        Gets a specific app revision.

        Supports resource cache

        Keyword Args:
            revision_number (float): App revision number

        Returns:
            AppRevision: The AppRevision for the given revision number.
        """
        key_value = AppRevision.get_unique_id(self._app.id, revision_number)
        return self.__get(app_id_revision=key_value)

    @check_cache(AppRevision)
    @one_of_keyword_only('app_id_revision')
    def __get(self, key, value):
        """Underlying get method supporting resource cache."""
        app_id, revision_number = AppRevision.parse_unique_id(value)
        app_revision_raw = self._swimlane.request('get', 'app/{0}/history/{1}'.format(app_id, revision_number)).json()
        return AppRevision(self._swimlane, app_revision_raw)



"""Provides handful of constants used in record reports/searches"""

# Filter
EQ = "equals"
NOT_EQ = "doesNotEqual"
CONTAINS = "contains"
EXCLUDES = "excludes"
GT = "greaterThan"
LT = "lessThan"
LTE = "lessThanOrEqual"
GTE = "greaterThanOrEqual"

# Aggregate
AVG = "average"
COUNT = "count"
SUM = "sum"
MIN = "min"
MAX = "max"

# GroupBy
GB = "groupBy"
HOUR = "groupByHour"
DAY = "groupByDay"
WEEK = "groupByWeek"
MONTH = "groupByMonth"
QUARTER = "groupByQuarter"
YEAR = "groupByYear"

# Sorts
ASC = "ascending"
DESC = "descending"



"""Core Swimlane client class"""

import logging

import jwt
import pendulum
import requests
from pyuri import URI
from requests.compat import json
from requests.packages import urllib3
from requests.structures import CaseInsensitiveDict
from six.moves.urllib.parse import urljoin

from swimlane.core.adapters import GroupAdapter, UserAdapter, AppAdapter, HelperAdapter
from swimlane.core.cache import ResourcesCache
from swimlane.core.resolver import SwimlaneResolver
from swimlane.core.resources.usergroup import User
from swimlane.exceptions import SwimlaneHTTP400Error, InvalidSwimlaneProductVersion
from swimlane.utils.version import get_package_version, compare_versions
from swimlane.core.wrappedsession import WrappedSession

# Disable insecure request warnings
urllib3.disable_warnings()

logger = logging.getLogger(__name__)

# pylint: disable=invalid-name
_lib_full_version = get_package_version()
_lib_major_version, _lib_minor_version = _lib_full_version.split('.')[0:2]


class Swimlane(object):
    """Swimlane API client

    Core class used throughout library for all API requests and server interactions

    Args:
        host (str): Full RFC-1738 URL pointing to Swimlane host. Defaults will be provided for all parts
        username (str): Authentication username
        password (str): Authentication password
        verify_ssl (bool): Verify SSL (ignored on HTTP). Disable to use self-signed certificates
        default_timeout (int): Default request connect and read timeout in seconds for all requests
        verify_server_version (bool): Verify server version has same major version as client package. May require
            additional requests, set False to disable check
        resource_cache_size (int): Maximum number of each resource type to keep in memory cache. Set 0 to disable
            caching. Disabled by default
        access_token (str): Authentication token, used in lieu of a username and password
        write_to_read_only (bool): Enable the ability to write to Read-only fields

    Attributes:
        host (pyuri.URI): Full RFC-1738 URL pointing to Swimlane host
        apps (AppAdapter): :class:`~swimlane.core.adapters.app.AppAdapter` configured for current Swimlane instance
        users (UserAdapter): :class:`~swimlane.core.adapters.usergroup.UserAdapter` configured for current
            Swimlane instance
        groups (GroupAdapter): :class:`~swimlane.core.adapters.usergroup.GroupAdapter` configured for current
            Swimlane instance
        resources_cache (ResourcesCache): Cache checked by all supported adapters for current Swimlane instance

    Examples:

        ::

            # Establish connection using username password
            swimlane = Swimlane(
                'https://192.168.1.1',
                'username',
                'password',
                verify_ssl=False
            )

            # Or establish connection using personal access token
            swimlane = Swimlane(
                'https://192.168.1.1',
                access_token='abcdefg',
                verify_ssl=False
            )

            # Retrieve an app
            app = swimlane.apps.get(name='Target App')

    """

    _api_root = '/api/'

    def __init__(
            self,
            host,
            username=None,
            password=None,
            verify_ssl=True,
            default_timeout=60,
            verify_server_version=True,
            resource_cache_size=0,
            access_token=None,
            write_to_read_only=False
    ):
        self.__verify_auth_params(username, password, access_token)

        self.host = URI(host)
        self.host.scheme = (self.host.scheme or 'https').lower()
        self.host.path = None

        self.resources_cache = ResourcesCache(resource_cache_size)

        self.__settings = None
        self.__user = None

        self._write_to_read_only = write_to_read_only

        self._default_timeout = default_timeout

        self._session = WrappedSession()
        self._session.verify = verify_ssl

        if username is not None and password is not None:
            self._session.auth = SwimlaneJwtAuth(
                self,
                username,
                password
            )
        else:
            self._session.auth = SwimlaneTokenAuth(
                self,
                access_token
            )

        self.apps = AppAdapter(self)
        self.users = UserAdapter(self)
        self.groups = GroupAdapter(self)
        self.helpers = HelperAdapter(self)

        if verify_server_version:
            self.__verify_server_version()

    @staticmethod
    def __verify_auth_params(username, password, access_token):
        """Verify that valid authentication parameters were passed to __init__"""

        if all(v is not None for v in [username, password, access_token]):
            raise ValueError('Cannot supply a username/password and a access token')

        if (username is None or password is None) and access_token is None:
            raise ValueError('Must supply a username/password or access token')

    def __verify_server_version(self):
        """Verify connected to supported server product version

        Notes:
            Logs warning if connecting to a newer minor server version

        Raises:
            swimlane.exceptions.InvalidServerVersion: If server major version is higher than package major version
        """
        if compare_versions('.'.join([_lib_major_version, _lib_minor_version]), self.product_version) > 0:
            logger.warning('Client version {} connecting to server with newer minor release {}.'.format(
                _lib_full_version,
                self.product_version
            ))

        if compare_versions(_lib_major_version, self.product_version) != 0:
            raise InvalidSwimlaneProductVersion(
                self,
                '{}.0'.format(_lib_major_version),
                '{}.0'.format(str(int(_lib_major_version) + 1))
            )

    def __repr__(self):
        return '<{cls}: {user} @ {host} v{version}>'.format(
            cls=self.__class__.__name__,
            user=self.user,
            host=self.host,
            version=self.version
        )

    def request(self, method, api_endpoint, **kwargs):
        """Wrapper for underlying :class:`requests.Session`

        Handles generating full API URL, session reuse and auth, request defaults, and invalid response status codes

        Used throughout library as the core underlying request/response method for all interactions with server

        Args:
            method (str): Request method (get, post, put, etc.)
            api_endpoint (str): Portion of URL matching API endpoint route as listed in platform /docs help page
            **kwargs (dict): Remaining arguments passed through to actual request call

        Notes:
            All other provided kwargs are passed to underlying :meth:`requests.Session.request()` call

        Raises:
            swimlane.exceptions.SwimlaneHTTP400Error: On 400 responses with additional context about the exception
            requests.HTTPError: Any other 4xx/5xx HTTP responses

        Returns:
            requests.Response: Successful response instances

        Examples:

            Request and parse server settings endpoint response

            >>> server_settings = swimlane.request('get', 'settings').json()
        """
        while api_endpoint.startswith('/'):
            api_endpoint = api_endpoint[1:]

        # Ensure a timeout is set
        kwargs.setdefault('timeout', self._default_timeout)

        # Manually grab and dump json data to have full control over serialization
        # Emulate default requests behavior
        json_data = kwargs.pop('json', None)
        if json_data is not None:
            headers = CaseInsensitiveDict(kwargs.get('headers', {}))
            headers.setdefault('Content-Type', 'application/json')
            kwargs['headers'] = headers

            kwargs['data'] = json.dumps(json_data, sort_keys=True, separators=(',', ':'))

        response = self._session.request(method, urljoin(str(self.host) + self._api_root, api_endpoint), **kwargs)

        # Roll 400 errors up into SwimlaneHTTP400Errors with specific Swimlane error code support
        try:
            response.raise_for_status()
        except requests.HTTPError as error:
            if error.response.status_code == 400:
                raise SwimlaneHTTP400Error(error)
            else:
                raise error

        return response

    @property
    def settings(self):
        """Retrieve and cache settings from server"""
        if not self.__settings:
            self.__settings = self.request('get', 'settings').json()
        return self.__settings

    @property
    def version(self):
        """Full Swimlane version, <product_version>+<build_version>+<build_number>"""
        return self.settings['apiVersion']

    @property
    def product_version(self):
        """Swimlane product version"""
        version_separator = '+'
        if version_separator in self.version:
            # Post product/build version separation
            return self.version.split(version_separator)[0]
        # Pre product/build version separation
        return self.version.split('-')[0]

    @property
    def build_version(self):
        """Swimlane semantic build version

        Falls back to product version in pre-2.18 releases
        """
        version_separator = '+'
        if version_separator in self.version:
            # Post product/build version separation
            # This will handle <product_version>+<build_version>+<build_number>
            # or <build_version>+<build_number> formats of the version
            return self.version.split(version_separator)[-2]
        # Pre product/build version separation
        return self.product_version

    @property
    def build_number(self):
        """Swimlane build number"""
        version_separator = '+'
        if version_separator in self.version:
            # Post product/build version separation
            return self.version.split(version_separator)[2]
        # Pre product/build version separation
        return self.version.split('-')[1]

    @property
    def user(self):
        """User record instance for authenticated user"""
        return self._session.auth.user


class SwimlaneTokenAuth(SwimlaneResolver):
    """Handles token authentication for all requests

    .. versionadded:: 4.1.0
    """

    def __init__(self, swimlane, access_token):
        super(SwimlaneTokenAuth, self).__init__(swimlane)

        self._access_token = access_token
        self.user = None
    
    def __call__(self, request):
        """Attach necessary headers to all requests"""

        headers = {
            'Private-Token': self._access_token
        }

        request.headers.update(headers)

        # Only make the call to user/authorize to get the user's profile if we haven't retrieved it
        # already
        if self.user is not None:
            return request

        # Temporarily remove auth from Swimlane session for auth request to avoid recursive loop during the request
        self._swimlane._session.auth = None
        resp = self._swimlane.request(
            'get',
            'user/authorize',
            headers=headers
        )
        self._swimlane._session.auth = self
        
        json_content = resp.json()
        self.user = User(self._swimlane, _user_raw_from_login_content(json_content))
        
        return request


class SwimlaneJwtAuth(SwimlaneResolver):
    """Handles authentication for all requests"""

    _token_expiration_buffer = pendulum.Duration(minutes=5)

    def __init__(self, swimlane, username, password):
        super(SwimlaneJwtAuth, self).__init__(swimlane)

        self._username = username
        self._password = password

        self.user = None
        self._login_headers = {}
        self._token_expiration = pendulum.now()

    def __call__(self, request):
        """Attach necessary headers to all requests

        Automatically reauthenticate before sending request when nearing token expiration
        """

        # Refresh token if it expires soon
        if pendulum.now() + self._token_expiration_buffer >= self._token_expiration:
            self.authenticate()

        request.headers.update(self._login_headers)

        return request

    def authenticate(self):
        """Send login request and update User instance, login headers, and token expiration"""

        # Temporarily remove auth from Swimlane session for auth request to avoid recursive loop during login request
        self._swimlane._session.auth = None
        resp = self._swimlane.request(
            'post',
            'user/login',
            json={
                'userName': self._username,
                'password': self._password
            },
        )
        self._swimlane._session.auth = self

        # Get JWT from response content
        json_content = resp.json()
        token = json_content.pop('token', None)

        # Grab token expiration
        token_data = jwt.decode(token, verify=False)
        token_expiration = pendulum.from_timestamp(token_data['exp'])

        headers = {
            'Authorization': 'Bearer {}'.format(token)
        }

        # Create User instance for authenticating user from login response data
        user = User(self._swimlane, _user_raw_from_login_content(json_content))

        self._login_headers = headers
        self.user = user
        self._token_expiration = token_expiration


def _user_raw_from_login_content(login_content):
    """Returns a User instance with appropriate raw data parsed from login response content"""
    matching_keys = [
        'displayName',
        'lastLogin',
        'active',
        'name',
        'isMe',
        'lastPasswordChangedDate',
        'passwordResetRequired',
        'groups',
        'roles',
        'email',
        'isAdmin',
        'createdDate',
        'modifiedDate',
        'createdByUser',
        'modifiedByUser',
        'userName',
        'id',
        'disabled'
    ]

    raw_data = {
        '$type': User._type,
    }

    for key in matching_keys:
        if key in login_content:
            raw_data[key] = login_content[key]

    return raw_data



import pendulum

from swimlane.core.resolver import SwimlaneResolver
from swimlane.utils.version import requires_swimlane_version


class HelperAdapter(SwimlaneResolver):
    """Adapter providing any miscellaneous API calls not better suited for another adapter"""

    @requires_swimlane_version('2.15')
    def add_record_references(self, app_id, record_id, field_id, target_record_ids):
        """Bulk operation to directly add record references without making any additional requests

        Warnings:
            Does not perform any app, record, or target app/record validation

        Args:
            app_id (str): Full App ID string
            record_id (str): Full parent Record ID string
            field_id (str): Full field ID to target reference field on parent Record string
            target_record_ids (List(str)): List of full target reference Record ID strings
        """

        self._swimlane.request(
            'post',
            'app/{0}/record/{1}/add-references'.format(app_id, record_id),
            json={
                'fieldId': field_id,
                'targetRecordIds': target_record_ids
            }
        )

    def add_comment(self, app_id, record_id, field_id, message, rich_text=False):
        """Directly add a comment to a record without retrieving the app or record first

        Warnings:
            Does not perform any app, record, or field ID validation

        Args:
            app_id (str): Full App ID string
            record_id (str): Full parent Record ID string
            field_id (str): Full field ID to target reference field on parent Record string
            message (str): New comment message body
            rich_text (bool): Declare the message as being rich text, default is False
        """
        if not isinstance(rich_text, bool):
            raise ValueError("rich_text must be a boolean value.")

        self._swimlane.request(
            'post',
            'app/{0}/record/{1}/{2}/comment'.format(
                app_id,
                record_id,
                field_id
            ),
            json={
                'message': message,
                'isRichText': rich_text,
                'createdDate': pendulum.now().to_rfc3339_string()
            }
        )

    def check_bulk_job_status(self, job_id):
        """Check status of bulk_delete or bulk_modify jobs
        .. versionadded:: 2.17.0
        Args:
            job_id (str): Job ID

        Returns:
            :class:`list` of :class:`dict`: List of dictionaries containing job history

        """

        return self._swimlane.request('get', "logging/job/{0}".format(job_id)).json()



"""Contains various adapters encapsulating API logic for retrieving, searching, listing, or creating resource objects"""

from .app import AppAdapter
from .record import RecordAdapter
from .report import ReportAdapter
from .usergroup import UserAdapter, GroupAdapter
from .helper import HelperAdapter
from .app_revision import AppRevisionAdapter
from .record_revision import RecordRevisionAdapter
from .task import TaskAdapter



from six.moves.urllib.parse import quote_plus

from swimlane.core.cache import check_cache
from swimlane.core.cursor import PaginatedCursor
from swimlane.core.resolver import SwimlaneResolver
from swimlane.core.resources.usergroup import Group, User
from swimlane.utils import one_of_keyword_only


class GroupListCursor(SwimlaneResolver, PaginatedCursor):
    """Handles retrieval and pagination of group list endpoint"""

    def __init__(self, swimlane, limit=None):
        SwimlaneResolver.__init__(self, swimlane)
        PaginatedCursor.__init__(self, limit)

    def _parse_raw_element(self, raw_element):
        return Group(self._swimlane, raw_element)

    def _retrieve_raw_elements(self, page):
        response = self._swimlane.request(
            'get',
            'groups',
            params={
                'size': self.page_size,
                'pageNumber': page
            }
        )
        return response.json().get('items', [])


class GroupAdapter(SwimlaneResolver):
    """Handles retrieval of Swimlane Group resources"""

    def list(self, limit=None):
        """Retrieve list of all groups

        Returns:
            :class:`list` of :class:`~swimlane.core.resources.usergroup.Group`: List of all Groups
        """
        return GroupListCursor(self._swimlane, limit=limit)

    @check_cache(Group)
    @one_of_keyword_only('id', 'name')
    def get(self, key, value):
        """Retrieve single group record by id or name

        Supports resource cache

        Keyword Args:
            id (str): Full Group ID
            name (str): Group name

        Raises:
            TypeError: Unexpected or more than one keyword argument provided
            ValueError: No matching group found based on provided inputs

        Returns:
            Group: Group instance matching provided inputs
        """
        if key == 'id':
            response = self._swimlane.request('get', 'groups/{}'.format(value))
            return Group(self._swimlane, response.json())

        else:
            response = self._swimlane.request('get', 'groups/lookup?name={}'.format(value))
            matched_groups = response.json()

            for group_data in matched_groups:
                if group_data.get('name') == value:
                    return Group(self._swimlane, group_data)

            raise ValueError('Unable to find group with name "{}"'.format(value))


class UserListCursor(SwimlaneResolver, PaginatedCursor):
    """Handles retrieval and pagination for user list endpoint"""

    def __init__(self, swimlane, limit=None):
        SwimlaneResolver.__init__(self, swimlane)
        PaginatedCursor.__init__(self, limit)

    def _parse_raw_element(self, raw_element):
        return User(self._swimlane, raw_element)

    def _retrieve_raw_elements(self, page):
        response = self._swimlane.request(
            'get',
            'user',
            params={
                'size': self.page_size,
                'pageNumber': page
            }
        )
        return response.json().get('items', [])


class UserAdapter(SwimlaneResolver):
    """Handles retrieval of Swimlane User resources"""

    def list(self, limit=None):
        """Retrieve all users

        Returns:
            :class:`UserListCursor`: Paginated cursor yielding :class:`User` instances
        """
        return UserListCursor(swimlane=self._swimlane, limit=limit)

    @check_cache(User)
    @one_of_keyword_only('id', 'display_name')
    def get(self, arg, value):
        """Retrieve single user record by id or username

        Warnings:
            User display names are not unique. If using `display_name`, method will fail if multiple Users are returned
            with the same display name

        Keyword Args:
            id (str): Full User ID
            display_name (str): User display name

        Returns:
            User: User instance matching provided inputs

        Raises:
            TypeError: Unexpected or more than one keyword argument provided
            ValueError: No matching user found based on provided inputs, or multiple Users with same display name
        """
        if arg == 'id':
            response = self._swimlane.request('get', 'user/{}'.format(value))

            try:
                user_data = response.json()
            except ValueError:
                raise ValueError('Unable to find user with ID "{}"'.format(value))

            return User(self._swimlane, user_data)

        else:
            response = self._swimlane.request('get', 'user/search?query={}'.format(quote_plus(value)))
            matched_users = response.json()

            # Display name not unique, fail if multiple users share the same target display name
            target_matches = []

            for user_data in matched_users:
                user_display_name = user_data.get('displayName')
                if user_display_name == value:
                    target_matches.append(user_data)

            # No matches
            if not target_matches:
                raise ValueError('Unable to find user with display name "{}"'.format(value))

            # Multiple matches
            if len(target_matches) > 1:
                raise ValueError('Multiple users returned with display name "{}". Matching user IDs: {}'.format(
                    value,
                    ', '.join(['"{}"'.format(r['id']) for r in target_matches])
                ))

            return User(self._swimlane, target_matches[0])



import weakref

from swimlane.core.resolver import SwimlaneResolver
from swimlane.core.resources.report import Report, report_factory


class ReportAdapter(SwimlaneResolver):
    """Handles retrieval and creation of Report resources"""

    def __init__(self, app):
        super(ReportAdapter, self).__init__(app._swimlane)

        self.__ref_app = weakref.ref(app)

    @property
    def _app(self):
        """Resolve weak app reference"""
        return self.__ref_app()

    def list(self):
        """Retrieve all reports for parent app

        Returns:
            :class:`list` of :class:`~swimlane.core.resources.report.Report`: List of all returned reports
        """
        raw_reports = self._swimlane.request('get', "reports/app/{}".format(self._app.id)).json()
        # Ignore StatsReports for now
        return [Report(self._app, raw_report) for raw_report in raw_reports if raw_report['$type'] == Report._type]

    def get(self, report_id):
        """Retrieve report by ID

        Args:
            report_id (str): Full report ID

        Returns:
            Report: Corresponding Report instance
        """
        return Report(
            self._app,
            self._swimlane.request('get', "reports/{0}".format(report_id)).json()
        )

    def build(self, name, **kwargs):
        """Report instance factory for the adapter's App

        Args:
            name (str): New Report name

        Keyword Args:
            **kwargs: Extra keyword args passed to Report class

        Returns:
            Report: Newly created local Report instance
        """
        return report_factory(self._app, name, **kwargs)



from swimlane.core.resources.base import APIResource


class Task(APIResource):
    _type = 'Core.Models.Integrations.Task, Core'

    def __init__(self, swimlane, raw):
        super(Task, self).__init__(swimlane, raw)
        self.app_id = raw.get('applicationId')
        self.id = raw.get('id')
        self.name = raw.get('name')
        self.script = raw.get('action').get('script')
    
    def __str__(self):
        return '{self.name}'.format(self=self)




import pendulum

from swimlane.core.resources.comment import Comment
from .base import CursorField, FieldCursor, ReadOnly


class CommentCursor(FieldCursor):
    """Returned by CommentField to allow iteration and creation of Comment instances"""

    def comment(self, message, rich_text=False):
        """Add new comment to record comment field"""
        message = str(message)
        if not isinstance(rich_text, bool):
            raise ValueError("rich_text must be a boolean value.")

        sw_repr = {
            '$type': 'Core.Models.Record.Comments, Core',
            'createdByUser': self._record._swimlane.user.as_usergroup_selection(),
            'createdDate': pendulum.now().to_rfc3339_string(),
            'message': message,
            'isRichText': rich_text
        }

        comment = Comment(self._swimlane, sw_repr)
        self._elements.append(comment)

        self._record._raw['comments'].setdefault(self._field.id, [])
        self._record._raw['comments'][self._field.id].append(comment._raw)

        # Tracking comment changes for patch endpoint
        self._record._comments_modified = True

        return comment


class CommentsField(ReadOnly, CursorField):

    field_type = (
        'Core.Models.Fields.CommentsField, Core',
        'Core.Models.Fields.Comments.CommentsField, Core'
    )
    cursor_class = CommentCursor
    bulk_modify_support = False

    def get_initial_elements(self):
        raw_comments = self.record._raw['comments'].get(self.id, [])

        return [Comment(self.record._swimlane, raw) for raw in raw_comments]



from swimlane.core.resources.record import Record
from swimlane.core.resources.revision_base import RevisionBase


class RecordRevision(RevisionBase):
    """
    Encapsulates a single revision returned from a History lookup.

    Attributes:
        app_revision_number: The app revision number this record revision was created using.

    Properties:
        app_version: Returns an App corresponding to the app_revision_number of this record revision.
        version: Returns a Record corresponding to the app_version and data contained in this record revision.
    """
    def __init__(self, app, raw):
        super(RecordRevision, self).__init__(app._swimlane, raw)

        self.__app_version = None
        self._app = app

        self.app_revision_number = self._raw_version['applicationRevision']

    @property
    def app_version(self):
        """The app revision corresponding to this record revision. Lazy loaded"""
        if not self.__app_version:
            self.__app_version = self._app.revisions.get(self.app_revision_number).version
        return self.__app_version

    @property
    def version(self):
        """The record contained in this record revision. Lazy loaded. Overridden from base class."""
        if not self._version:
            self._version = Record(self.app_version, self._raw_version)
        return self._version
